import imp
import os

plugin_folder = os.path.join(os.path.dirname(os.path.abspath(__file__)))
main_module = '__init__'


def get_plugins():
    plugins = []
    possible_plugins = os.listdir(plugin_folder)
    for i in possible_plugins:
        location = os.path.join(plugin_folder, i)
        if not os.path.isdir(location) or not main_module + '.py' in os.listdir(location):
            continue
        info = imp.find_module(main_module, [location])
        plugins.append({'name': i, 'info': info})
    return plugins


def load_plugin(plugin):
    return imp.load_module(plugin['name'], *plugin['info'])



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
# A tool to help manage Xonotic maps
# z@xnz.me

import argcomplete
import argparse
import logging
import os

from xmm import __version__
from xmm.config import conf

from xmm.server import LocalServer
from xmm.server import ServerCollection

from xmm.exceptions import HashMismatchError
from xmm.exceptions import PackageMetadataWarning
from xmm.exceptions import PackageNotTrackedWarning
from xmm.exceptions import PackageLookupError
from xmm.exceptions import RepositoryLookupError
from xmm.exceptions import RepositoryUpdateError
from xmm.exceptions import ServerLookupError
from xmm.plugins import pluginbase
from xmm.plugins import pluginloader
from xmm.logger import ClassPrefixAdapter
from xmm.util import cprint
from xmm.util import zcolors
from xmm import util

plugins = {}

cli_logger = ClassPrefixAdapter(prefix='cli', logger=logging.getLogger(__name__))


def main():

    pluginbase.set_config(conf)
    args = parse_args()
    server = None

    # Just install
    if args.target:
        if args.command != 'install':
            cprint("This flag only works with 'install' sub-command.", style='FAIL')

        target_dir = args.target
        cli_logger.info("Installing to target directory: {}".format(target_dir))

        if not os.path.exists(target_dir):
            cli_logger.error("Target directory does not exist: {}.".format(target_dir))
            cprint("Target directory does not exist.", style='FAIL')
            raise SystemExit

        filename_with_path = os.path.join(target_dir, args.pk3)
        url_with_file = '{}/{}'.format(conf['default']['download_url'], args.pk3)

        success = util.download_file(filename_with_path=filename_with_path, url=url_with_file, use_curl=conf['default']['use_curl'])

        if not success:
            cli_logger.warning("{} already exists, not installing.".format(args.pk3))

        exit(0)

    # Use all source repositories
    else:
        try:
            server = LocalServer(server_name=args.server)
        except NotADirectoryError as e:
            create_directory = util.query_yes_no("Directory '{}' does not exist, would you like to create it?".format(e))
            if create_directory:
                server = LocalServer(server_name=args.server, make_dirs=True)
            else:
                cprint("Canceled.", style='INFO')
                raise SystemExit
        except ServerLookupError as e:
            cprint("server '{}' does not exist in ~/.xmm/servers.json".format(e), style='FAIL')

    # Sort out defaults
    if 'long' in args and args.long:
        detail = 'long'
    elif 'short' in args and args.short:
        detail = 'short'
    else:
        detail = None

    highlight = False
    if 'highlight' in args and args.highlight:
        highlight = True

    # Commands
    if args.command == 'search':

        try:
            server.repositories.search_all(bsp_name=args.string, gametype=args.gametype, author=args.author,
                                           title=args.title, pk3_name=args.pk3, shasum=args.shasum, detail=detail,
                                           highlight=highlight)
        except Exception:
            cprint('Failed.', style='FAIL')

    elif args.command == 'install':

        if not args.pk3:
            cprint("package name not specified", style='FAIL')
            raise SystemExit

        cprint("Installing map: {}".format(args.pk3), style='BOLD')

        try:
            server.library.install_map(pk3_name=args.pk3, repository_name=args.repository)
        except SystemExit:
            cprint("Canceled.", style='INFO')
        except NotADirectoryError as e:
            cprint("package directory does not exist: {}".format(e), style='FAIL')
        except PackageMetadataWarning:
            cprint("package does not exist in the repository it won't be tracked (xmm list).", style='WARNING')
        except RepositoryLookupError:
            cprint("Repository does not exist!", style='FAIL')
        except PackageLookupError:
            cprint("package does not exist in the repository. cannot install.", style='FAIL')

    elif args.command == 'remove':

        if not args.pk3:
            cprint("package name not specified", style='FAIL')
            raise SystemExit

        cprint("Removing package: {}".format(args.pk3), style='BOLD')

        try:
            server.library.remove_map(pk3_name=args.pk3)
            cprint("Done.", style='INFO')
        except FileNotFoundError:
            cprint("package does not exist or is not tracked. try removing using full path if not tracked.", style='FAIL')
        except NotADirectoryError as e:
            cprint("package directory does not exist: {}".format(e), style='FAIL')

    elif args.command == 'discover':

        repository_name = None

        if args.repository:
            repository_name = args.repository
            cprint("Using repo '{}'".format(args.repository), style='HEADER')
            try:
                repo = server.repositories.get_repository(repository_name)
            except RepositoryLookupError:
                cprint("Repository doesn't exist in sources.json", style="FAIL")
                raise SystemExit

        try:
            server.library.discover_maps(add=args.add, repository_name=repository_name, detail=detail)
        except NotADirectoryError as e:
            cprint("package directory does not exist: {}".format(e), style='FAIL')

    elif args.command == 'list':

        try:
            total = server.library.list_installed(detail=detail)
            print("\n{}Total packages found:{} {}{}{}".format(zcolors.INFO, zcolors.ENDC, zcolors.BOLD, str(total), zcolors.ENDC))
        except Exception:
            cprint("Failed.", style='FAIL')

    elif args.command == 'show':

        if not args.pk3:
            cprint("package name not specified", style='FAIL')
            raise SystemExit

        if not detail:
            detail = 'long'

        # Use local package store for lookup
        if args.local:

            try:
                server.library.show_map(pk3_name=args.pk3, detail=detail, highlight=highlight)
            except HashMismatchError:
                print("\n{}{}{} {}hash different from repositories{}".format(zcolors.BOLD, args.pk3, zcolors.ENDC, zcolors.WARNING, zcolors.ENDC))
            except PackageNotTrackedWarning:
                print("\n{}{}{} {}package not currently tracked{}".format(zcolors.BOLD, args.pk3, zcolors.ENDC, zcolors.WARNING, zcolors.ENDC))

        # Use repositories for lookup
        else:

            if args.repository:

                try:
                    repo = server.repositories.get_repository(args.repository)
                except RepositoryLookupError:
                    cprint("Repository doesn't exist in sources.json", style="FAIL")
                    raise SystemExit

                repo.show_map(pk3_name=args.pk3, detail=detail, highlight=highlight)

            else:

                try:
                    for repo in server.repositories.sources:
                        cprint("Using repo '{}'".format(repo.name), style='HEADER')
                        map_found = repo.show_map(pk3_name=args.pk3, detail=detail, highlight=highlight)
                        if map_found:
                            break
                except PackageLookupError:
                    cprint("Map was not found in repository", style="FAIL")

    elif args.command == 'export':

        repository_name = None

        if args.server and args.repository:
            cprint("'-R' and '-S' flags are mutually exclusive on this command.", style="FAIL")
            raise SystemExit

        if args.repository:
            cprint("Using repo '{}'".format(args.repository), style='HEADER')
            repository_name = args.repository

        if args.format == 'json':

            if args.subcommand == 'local':

                server.library.export_map_packages(filename=args.filename)

            elif args.subcommand == 'repos':

                if repository_name:
                    try:
                        server.library.repositories.get_repository(repository_name).export_packages(filename=args.filename)
                    except RepositoryLookupError:
                        cprint("Repository doesn't exist in sources.json", style="FAIL")
                        raise SystemExit
                else:
                    server.library.repositories.export_all_packages(filename=args.filename)

        if args.format == 'maplist':

            if args.subcommand == 'local':

                server.library.export_maplist(filename=args.filename)

            elif args.subcommand == 'repos':

                # TODO: implement
                print('sorry exporting bsp names from repos is not supported yet')
                exit(0)

        elif args.format == 'shasums':

            if args.subcommand == 'local':

                if not repository_name:
                    server.library.export_hash_index(filename=args.filename)
                else:
                    # TODO: limit export by repository name
                    cprint('this combination is not yet possible')

            elif args.subcommand == 'repos':

                if repository_name:
                    try:
                        server.library.repositories.get_repository(repository_name).export_hash_index(filename=args.filename)
                    except RepositoryLookupError:
                        cprint("Repository doesn't exist in sources.json", style="FAIL")
                        raise SystemExit
                else:
                    server.library.repositories.export_all_hash_index(filename=args.filename)

    elif args.command == 'servers':

        servers = ServerCollection(servers=[])

        if args.subcommand == 'list':
            servers.list_servers()

    elif args.command == 'repos':

        if args.subcommand == 'list':
            server.repositories.list_repositories()

    elif args.command == 'update':

        try:
            server.repositories.update_all()
        except RepositoryUpdateError:
            cprint('One or more repositories have failed to update.', style='FAIL')

    # Plugins
    for cmd, value in plugins.items():
        if args.command == cmd:
            plugins[cmd].run()
            break


def parse_args():

    global plugins

    parser = argparse.ArgumentParser(description='Xonotic Map Manager is a tool to help manage Xonotic maps')

    parser.add_argument('--version', action='version', version='%(prog)s {0}'.format(__version__))
    parser.add_argument("-S", '--server', nargs='?', help="target server as defined in servers.json", type=str)
    parser.add_argument("-T", '--target', nargs='?', help="target directory", type=str)
    parser.add_argument("-R", '--repository', nargs='?', help="repository to use (defaults to all available)", type=str, default=None)

    subparsers = parser.add_subparsers(dest='command')
    subparsers.required = True

    parser_search = subparsers.add_parser('search', help='search for maps based on bsp names')
    parser_search.add_argument('string', nargs='?', help='bsp name found in a package, works on packages with many bsps', type=str)
    parser_search.add_argument('--gametype', nargs='?', help='filter by gametype', type=str)
    parser_search.add_argument('--pk3', nargs='?', help='filter by pk3 name', type=str)
    parser_search.add_argument('--title', nargs='?', help='filter by title', type=str)
    parser_search.add_argument('--author', nargs='?', help='filter by author', type=str)
    parser_search.add_argument('--shasum', nargs='?', help='filter by shasum', type=str)
    parser_search.add_argument('--long', '-l', help='show long format', action='store_true')
    parser_search.add_argument('--short', '-s', help='show short format', action='store_true')
    parser_search.add_argument('--color', '-c', help='highlight search term in results', action='store_true')

    parser_install = subparsers.add_parser('install', help='install a map from the repository, or specify a URL.')
    parser_install.add_argument('pk3', nargs='?', help='use a pk3 name of map package, or specify a URL of a pk3.', type=str)

    parser_remove = subparsers.add_parser('remove', help='remove based on pk3 name')
    parser_remove.add_argument('pk3', nargs='?', help='pk3', type=str)

    parser_discover = subparsers.add_parser('discover', help='discover packages in a target directory')
    parser_discover.add_argument('--long', '-l', help='show long format', action='store_true')
    parser_discover.add_argument('--short', '-s', help='show short format', action='store_true')
    parser_discover.add_argument('--add', '-a', help='add discovered files to the db', action='store_true')

    parser_list = subparsers.add_parser('list', help='list locally installed packages')
    parser_list.add_argument('--long', '-l', help='show long format', action='store_true')
    parser_list.add_argument('--short', '-s', help='show short format', action='store_true')

    parser_show = subparsers.add_parser('show', help='show details of remote or locally installed packages')
    parser_show.add_argument('pk3', nargs='?', help='pk3 to show details for', type=str)
    parser_show.add_argument('--local', '-L', help='whether to show from local packages only or all repos', action='store_true')
    parser_show.add_argument('--long', '-l', help='show long format', action='store_true')
    parser_show.add_argument('--short', '-s', help='show short format', action='store_true')

    parser_export = subparsers.add_parser('export', help='export locally managed packages to a file')
    parser_export.add_argument('subcommand', choices=['local', 'repos'], help='what context to export?', default='local', type=str)
    parser_export.add_argument('filename', nargs='?', help='filename to export to', type=str)
    parser_export.add_argument('--format', '-f', choices=['json', 'shasums', 'maplist'], default='json')

    parser_servers = subparsers.add_parser('servers', help='subcommands on servers described in servers.json')
    parser_servers.add_argument('subcommand', choices=['list'], help='list all servers in servers.json', type=str)

    parser_servers = subparsers.add_parser('repos', help='subcommands on repos described in sources.json')
    parser_servers.add_argument('subcommand', choices=['list'], help='list all servers in sources.json', type=str)

    parser_update = subparsers.add_parser('update', help='update sources json')

    # Handle plugins
    for i in pluginloader.get_plugins():
        cli_logger.debug("Loading plugin: " + i["name"])
        command = i['name']
        plugin = pluginloader.load_plugin(i)
        plugin_args = [plugin.get_args()]
        plugins[command] = plugin

        # Import args from plugins
        for p in plugin_args:
            parser_plugin = subparsers.add_parser(p[0], **p[1])
            parser_plugin.add_argument(*p[2], **p[3])

    argcomplete.autocomplete(parser)

    return parser.parse_args()


if __name__ == "__main__":
    main()



import json
import os

from xmm.base import Base
from xmm.exceptions import ServerLookupError
from xmm.library import Library
from xmm.repository import Repository
from xmm.repository import Collection
from xmm.store import Store
from xmm import util


# TODO: fleet management
class ServerCollection(Base):
    """
    A *ServerCollection* is a group of *LocalServer* objects. Currently unused.
    """
    def __init__(self, servers):
        super().__init__()
        self.servers = servers

    def __repr__(self):
        return str(vars(self))

    def __json__(self):
        return self.servers

    def to_json(self):
        """
        :returns: A **JSON** encoded version of this object
        """
        return json.dumps(self.servers, cls=util.ObjectEncoder)

    def list_servers(self):
        """
        Prints a list of servers
        """
        for server in self.conf['servers']:
            print(server)


class LocalServer(Base):
    """This class sets up the *LocalServer* object

    During instantiation, new objects are created based on configuration.

    The hierarchy of these objects looks like:

    * ``LocalServer``

        * ``Library``

            * ``Store``

            * ``MapPackage``

                * ``Bsp``

        * ``Collection``

            * ``Repository``


    :param server_name:
        Used to reference the server by name
    :type server_name: ``str``

    :param source_name:
        If specified, the server will be associated with this one source
    :type source_name: ``str``

    :param make_dirs:
        If directories aren't found that are required by xmm on server init, optionally create them
    :type make_dirs: ``bool``

    :returns object: ``LocalServer``
        Commands are available off ``self.library``.

    :Example:

    >>> from xmm.server import LocalServer
    >>> server = LocalServer(server_name='myserver1')
    >>> print(server)
    """

    def __init__(self, server_name='default', source_name=None, make_dirs=False):
        super().__init__()

        self.logger.info('initializing LocalServer: {}'.format(server_name))

        map_dir = self.conf['default']['target_dir']

        package_store_file = os.path.expanduser(self.conf['default']['library'])

        if server_name and server_name != 'default':
            try:
                map_dir = self.conf['servers'][server_name]['target_dir']
            except KeyError:
                self.logger.error('Server not found: {}'.format(server_name))
                raise ServerLookupError(server_name)
            server_data = self.conf['servers']
            if server_name in server_data:
                package_store_file = os.path.expanduser(server_data[server_name]['library'])

        self.map_dir = os.path.expanduser(map_dir)

        if not os.path.exists(self.map_dir):
            if make_dirs:
                os.makedirs(self.map_dir)
            else:
                self.logger.error('Directory not found: {}'.format(self.map_dir))
                raise NotADirectoryError(self.map_dir)

        store = Store(package_store_file=package_store_file)

        self.repositories = Collection()

        if source_name:
            self.logger.info('Using source: {}'.format(source_name))
            one_repo = self.conf['sources'][source_name]
            repository = Repository(name=source_name,
                                    download_url=one_repo['download_url'],
                                    api_data_url=one_repo['api_data_url'],
                                    api_data_file=one_repo['api_data_file']
                                    )

            self.repositories.add_repository(repository)

        else:
            self.logger.info('Using multiple sources'.format(source_name))
            for source_name in self.conf['sources']:
                self.logger.info('Using source: {}'.format(source_name))
                repository = Repository(name=source_name,
                                        download_url=self.conf['sources'][source_name]['download_url'],
                                        api_data_url=self.conf['sources'][source_name]['api_data_url'],
                                        api_data_file=self.conf['sources'][source_name]['api_data_file']
                                        )

                self.repositories.add_repository(repository)

        self.library = Library(store=store, repositories=self.repositories, map_dir=self.map_dir)

    def __repr__(self):
        return str(vars(self))

    def __json__(self):
        return {
            'repositories': self.repositories,
            'library': self.library,
        }

    def to_json(self):
        """
        :returns: A **JSON** encoded version of this object
        """
        return json.dumps(self, cls=util.ObjectEncoder)



__all__ = [
    '__title__',
    '__package_name__',
    '__version__',
    '__summary__',
    '__keywords__',
    '__author__',
    '__email__',
    '__author__',
    '__license__',
    '__url__',
    '__copyright__',
]

__title__ = 'Xonotic Map Manager'
__package_name__ = 'xmm'
__version__ = '0.8.0'
__summary__ = 'A package manager for Xonotic.'
__keywords__ = ['xonotic', 'xonotic-servers', 'xonotic-maps', 'gaming', 'cli']
__author__ = 'Tyler Mulligan'
__email__ = 'z@xnz.me'
__license__ = 'MIT'
__url__ = 'https://www.github.com/z/xonotic-map-manager'
__copyright__ = 'Copyright 2016 Tyler Mulligan'



import os
import re
import json

from xmm.exceptions import PackageMetadataWarning
from xmm.exceptions import PackageNotTrackedWarning
from xmm.exceptions import PackageLookupError
from xmm.exceptions import RepositoryLookupError
from xmm.exceptions import HashMismatchError
from xmm.base import Base
from xmm.util import cprint
from xmm import util


class Library(Base):
    """
    A *Library* is a collection of *MapPackage* objects and commands for managing maps in the *Library*

    :param repositories:
        A *Collection* object with *Repository* objects
    :type repositories: ``Collection``

    :param store:
        A *Store* object for communicate with the data store for this *Library*
    :type store: ``Store``

    :param map_dir:
        The directory this *Library* is associated with
    :type map_dir: ``str``

    :returns object: ``Library``

    """
    def __init__(self, repositories, store, map_dir):
        super().__init__()
        self.maps = store.get_package_db()
        self.repositories = repositories
        self.store = store
        self.map_dir = os.path.expanduser(map_dir)

    def __repr__(self):
        return str(vars(self))

    def __json__(self):
        return {
            'maps': self.maps,
            'repositories': self.repositories,
            'store': self.store,
            'map_dir': self.map_dir,
        }

    def to_json(self):
        """
        :returns: A **JSON** encoded version of this object
        """
        return json.dumps(self, cls=util.ObjectEncoder)

    def add_map_package(self, package):
        """
        Adds a *MapPackage* object to ``self.maps``

        :param package:
            A *MapPackage* object for the *Library*
        :type package: ``MapPackage``
        """
        self.maps.append(package)

    def get_repository_sources(self, server_name):
        """
        Gets the *Collection* from the *Library* of the specified *LocalServer* from ``self.repositories``
        as cache, or from the ``sources.json`` targeted by ``servers.json`` if it is not already set.

        :param server_name:
            Server name
        :type server_name: ``str``

        :returns: ``Collection``
        """
        if server_name:
            repo_sources = self.conf['servers'][server_name]['sources']

            with open(repo_sources) as f:
                data = f.read()
                sources = json.loads(data)
                self.repositories = sources
        else:
            repo_sources = self.repositories

        return repo_sources

    def install_map(self, pk3_name, repository_name=None, overwrite=False, add_to_store=True):
        """
        Install a *MapPackage* from a *Repository*

        :param pk3_name:
            A pk3 name such as ``vinegar_v3.pk3``, to install from the repository.'
            Optionally prefixed with a URL to install map not in the repository.
            URL-only maps will not include rich metadata available to maps installed via the repo.
        :type pk3_name: ``str``

        :param repository_name:
            A name of a repository in the repository *Collection*
        :type repository_name: ``str``

        :param overwrite:
            Whether to overwrite the file on the file system
        :type overwrite: ``bool``

        :param add_to_store:
            Whether to add the map to the store (tracked)
        :type add_to_store: ``bool``

        >>> from xmm.server import LocalServer
        >>> server = LocalServer(server_name='myserver1')
        >>> server.library.install_map(pk3_name='vinegar_v3.pk3')
        >>> print(server.library.maps)
        """
        self.logger.info('Installing map: {}.'.format(pk3_name))

        map_dir = self.map_dir
        installed_packages = self.store.get_package_db()
        map_found_in_repo = False
        found_map = None
        installed = False
        is_url = False

        if not os.path.exists(map_dir):
            raise NotADirectoryError(map_dir)

        if repository_name:
            repo = self.repositories.get_repository(repository_name)
            if repo:
                sources = [repo]
            else:
                raise RepositoryLookupError
        else:
            sources = self.repositories.sources

        if installed_packages and not overwrite:
            for m in installed_packages:
                if m.pk3_file == pk3_name:
                    cprint("{} already exists.".format(pk3_name), style='WARNING')
                    self.logger.warning("{} already exists.".format(pk3_name))
                    install = util.query_yes_no('continue?', 'no')
                    if not install:
                        raise SystemExit
                    else:
                        self.logger.info("overwriting {}.".format(pk3_name))
                        overwrite = True
                        add_to_store = False
                    break

        if re.match('^(ht|f)tp(s)?://', pk3_name):
            self.logger.info("{} downloading from non-repository link.".format(pk3_name))
            url = pk3_name
            pk3 = os.path.basename(pk3_name)
            is_url = True
            overwrite = False
        else:
            pk3 = pk3_name
            url = self.conf['default']['download_url'] + pk3

        pk3_with_path = os.path.join(os.path.dirname(map_dir), pk3)

        for repo in sources:
            if not map_found_in_repo:
                maps_json = repo.get_packages()
                for m in maps_json:
                    if m.pk3_file == pk3:
                        found_map = m
                        self.add_map_package(found_map)
                        map_found_in_repo = True
                        cprint("Found in: {}".format(repo.name))
                        break

        if map_found_in_repo or is_url:
            util.download_file(filename_with_path=pk3_with_path, url=url, use_curl=self.conf['default']['use_curl'], overwrite=overwrite)
            installed = True

        if installed and add_to_store and found_map:
            self.store.add_package(found_map)

        if not map_found_in_repo:
            if installed:
                self.logger.warning("{} was not installed through a repository and has no metadata.".format(pk3_name))
                raise PackageMetadataWarning
            else:
                self.logger.error("Unable to find package: {}".format(pk3_name))
                raise PackageLookupError

    def remove_map(self, pk3_name):
        """
        Removes a map from the *Library* and the package from the file system

        :param pk3_name:
            The name of a pk3, such as ``vinegar_v3.pk3``
        :type pk3_name: ``str``

        >>> from xmm.server import LocalServer
        >>> server = LocalServer(server_name='myserver1')
        >>> server.library.install_map_package(pk3_name='vinegar_v3.pk3')
        >>> print(server.library.maps)
        >>> server.library.remove_map(pk3_name='vinegar_v3.pk3')
        >>> print(server.library.maps)
        """

        self.logger.info("Removing map: {}".format(pk3_name))

        map_dir = os.path.expanduser(self.map_dir)
        pk3_with_path = os.path.join(os.path.dirname(map_dir), pk3_name)

        installed_packages = self.store.get_package_db()

        for m in installed_packages:
            if m.pk3_file == pk3_name:
                self.store.remove_package(m)

        if os.path.exists(pk3_with_path):
            os.remove(pk3_with_path)
        else:
            raise FileNotFoundError(pk3_with_path)

    def discover_maps(self, add=False, repository_name=None, detail=None):
        """
        Searches the *Server*'s map_dir for map packages known by the *Repository*

        :param add:
            Whether to add the discovered maps or not
        :type add: ``bool``

        :param repository_name:
            A name of a repository in the repository *Collection*
        :type repository_name: ``str``

        :param detail:
            How much detail to show, [short, None, long]
        :type detail: ``str``

        :param highlight:
            Whether to highlight the results
        :type highlight: ``bool``

        >>> from xmm.server import LocalServer
        >>> server = LocalServer()
        >>> server.library.discover_maps(add=False)
        """

        self.logger.debug("discovering maps")
        local_maps = self.store.get_package_db()

        if repository_name:
            repo = self.repositories.get_repository(repository_name)
            if repo:
                sources = [repo]
            else:
                self.logger.error("Unable to find repository in sources: {}".format(repo))
                raise RepositoryLookupError
        else:
            sources = self.repositories.sources

        if not os.path.exists(self.map_dir):
            self.logger.error("{} does not exists.".format(self.map_dir))
            raise NotADirectoryError(self.map_dir)

        for pk3_file in os.listdir(self.map_dir):

            map_found = False
            hash_match = False

            if pk3_file.endswith('.pk3'):
                shasum = util.hash_file(os.path.join(self.map_dir, pk3_file))

                try:
                    for repo in sources:
                        map_found = repo.show_map(pk3_file, detail=detail)
                        if map_found:
                            break
                except PackageLookupError:
                    pass

                if map_found:

                    if map_found.shasum == shasum:
                        hash_match = True
                    else:
                        self.logger.warning("{} hash does not match repository's".format(shasum))
                        cprint("{} hash does not match repository's".format(pk3_file), style='WARNING')

                    if hash_match and add:
                        map_already_installed = False
                        for m in local_maps:
                            if m.pk3_file == pk3_file and m.shasum == shasum:
                                self.logger.info("map already installed, not installing: {}".format(pk3_file))
                                map_already_installed = True

                        if not map_already_installed:
                            self.logger.info("installing map: {}".format(pk3_file))
                            self.store.add_package(map_found)

    # local data
    def list_installed(self, detail=None):
        """
        List maps currently tracked by the *Library*

        :param detail:
            How much detail to show, [short, None, long]
        :type detail: ``str``

        :returns: ``int`` total count

        >>> from xmm.server import LocalServer
        >>> server = LocalServer()
        >>> server.library.list_installed()
        """

        self.logger.debug("listing maps")

        packages = self.store.data

        total = 0
        if packages:
            for m in packages:
                m.show_map_details(detail=detail)
                total += 1

        return total

    def show_map(self, pk3_name, detail=None, highlight=False):
        """
        Convenience function to use the show_map_details helper

        :param pk3_name:
            The name of a pk3, such as ``vinegar_v3.pk3``
        :type pk3_name: ``str``

        :param detail:
            How much detail to show, [short, None, long]
        :type detail: ``str``

        :param highlight:
            Whether to highlight the results
        :type highlight: ``bool``

        :returns: ``MapPackage``

        >>> from xmm.server import LocalServer
        >>> server = LocalServer()
        >>> server.library.show_map('vinegar_v3.pk3', detail='long')
        """
        self.logger.debug("showing map: {}".format(pk3_name))

        packages = self.store.get_package_db()

        found_map = False
        hash_match = False

        for p in packages:
            if p.pk3_file == pk3_name:
                shasum = util.hash_file(os.path.join(self.map_dir, pk3_name))
                if p.shasum == shasum:
                    hash_match = True
                    p.show_map_details(search_string=pk3_name, detail=detail, highlight=highlight)
                    found_map = p
                else:
                    self.logger.warning("Hash for this map does not match repository's: {}".format(pk3_name))
                    raise HashMismatchError

        if not found_map and not hash_match:
            self.logger.warning("Map is not being tracked in the library: {}.".format(pk3_name))
            raise PackageNotTrackedWarning

        return found_map

    def export_map_packages(self, filename=None):
        """
        Exports all *MapPackage* objects from the *Library* *Store*

        :param filename:
            Name for the exported json file, default ``xmm-export.json``
        :type filename: ``str``

        :returns: False if fails

        >>> from xmm.server import LocalServer
        >>> # Setup the store automatically with an instance of *LocalServer*
        >>> server = LocalServer()
        >>> server.library.export_packages(filename='test.maps.json')
        """
        if not filename:
            filename = 'xmm-export.maps.json'

        data_out = []
        for m in self.maps:
            data_out.append(json.loads(m.to_json()))

        self.logger.info('exporting maps as: {}'.format(filename))

        try:
            with open(filename, 'w') as f:
                f.write(json.dumps(data_out))
        except EnvironmentError as e:
            self.logger.error(e)
            return False

    def export_hash_index(self, filename=None):
        """
        :param filename:
            Name for the exported json file, default ``maps.json.shasums``
        :type filename: ``str``

        :returns: False if fails

        >>> from xmm.server import LocalServer
        >>> # Setup the store automatically with an instance of *LocalServer*
        >>> server = LocalServer()
        >>> server.library.export_hash_index(filename='test.maps.shasums')
        """
        if not filename:
            filename = 'xmm-export.maps.shasums'

        self.logger.info("exporting shasums from all sources to file: {}".format(filename))

        data_out = []
        for m in self.maps:
            data_out.append("{} {}".format(m.shasum, m.pk3_file))

        if data_out:
            try:
                with open(filename, 'w') as f:
                    f.write('\n'.join(data_out))
            except EnvironmentError as e:
                self.logger.error(e)
                return False

    def export_maplist(self, filename=None):
        """
        :param filename:
            Name for the exported text file, default ``xmm-export.maps.txt``
        :type filename: ``str``

        :returns: False if fails

        >>> from xmm.server import LocalServer
        >>> # Setup the store automatically with an instance of *LocalServer*
        >>> server = LocalServer()
        >>> server.library.export_hash_index(filename='test.maps.txt')
        """
        if not filename:
            filename = 'xmm-export.maps.txt'

        self.logger.info("exporting maplist to file: {}".format(filename))

        data = []
        for m in self.maps:
            bsps = m.bsp
            bsp_names = list(bsps)
            bsp_names.sort()
            data.extend(bsp_names)

        if data:
            try:
                with open(filename, 'w') as f:
                    f.write('\n'.join(data))
            except EnvironmentError as e:
                self.logger.error(e)
                return False



from urllib.error import URLError


class HashMismatchError(ValueError):
    """
    Raise when a file hash mismatches
    """


class PackageLookupError(LookupError):
    """
    Raise when Package does not exist in Repository
    """


class PackageMetadataWarning(Warning):
    """
    Raise when Package installs from a URL and has no metadata associated with it.
    """


class PackageNotTrackedWarning(Warning):
    """
    Raise when Package is not tracked in the local db
    """


class RepositoryLookupError(LookupError):
    """
    Raise when Repository lookup fails
    """


class RepositoryUpdateError(URLError):
    """
    Raise when Repository update fails
    """


class ServerLookupError(LookupError):
    """
    Raise when Server lookup fails
    """



from .__about__ import (__title__, __package_name__, __version__, __summary__, __url__,
                        __keywords__, __email__, __author__, __license__, __copyright__)

from . import config
from . import logger

__all__ = [
    'config',
    'logger',
]



import logging
from xmm.config import conf
from xmm.logger import ClassPrefixAdapter


class Base(object):

    def __init__(self):
        self.conf = conf
        self.logger = ClassPrefixAdapter(prefix=self.__class__.__name__, logger=logging.getLogger(__name__))



import configparser
import os
import sys
import json
import time
import hashlib
import subprocess
import urllib.request
from datetime import datetime
from shutil import copyfile


def convert_size(number):
    """
    Convert and integer to a human-readable B/KB/MB/GB/TB string.

    :param number:
        integer to be converted to readable string
    :type number: ``int``

    :returns: `str`
    """
    for x in ['B', 'KB', 'MB', 'GB']:
        if number < 1024.0:
            string = "%3.1d%s" % (number, x)
            return string.strip()
        number /= 1024.0
    string = "%3.1f%s" % (number, 'TB')
    return string.strip()


def reporthook(count, block_size, total_size):
    """
    Pretty progress for urllib downloads.

    >>> import urllib.request
    >>> urllib.request.urlretrieve(url, filename, reporthook)

    https://github.com/yahoo/caffe/blob/master/scripts/download_model_binary.py
    """
    global start_time

    if count == 0:
        start_time = time.time()
        return
    duration = time.time() - start_time
    progress_size = int(count * block_size)
    speed = int(progress_size / (1024 * duration))
    percent = int(count * block_size * 100 / total_size)
    sys.stdout.write("\r...%d%%, %d MB, %d KB/s, %d seconds passed. " %
                    (percent, progress_size / (1024 * 1024), speed, duration))
    sys.stdout.flush()


def download_file(filename_with_path, url, use_curl=False, overwrite=False):
    """
    downloads a file from any URL

    :param filename_with_path:
        filename with path to download file to
    :type filename_with_path: ``str``

    :param url:
        URL to download map from
    :type url: ``str``

    :param use_curl:
        Whether or not to use curl to download the file, default ``False``

    :param overwrite:
        Whether or not to overwrite the existing file, default ``False``
    :type use_curl: ``bool``
    """
    if not os.path.exists(filename_with_path) or overwrite:

        if not use_curl:
            urllib.request.urlretrieve(url, os.path.expanduser(filename_with_path), reporthook)
        else:
            subprocess.call(['curl', '-o', filename_with_path, url])

        print("{}Done.{}".format(zcolors.INFO, zcolors.ENDC))

    else:
        print("{}file already exists, please remove first.{}".format(zcolors.FAIL, zcolors.ENDC))
        return False


def parse_config(config_file):
    """
    downloads a file from any URL

    :param config_file:
        filename with path to config file
    :type config_file: ``str``

    :returns: ``dict``
    """
    if not os.path.isfile(config_file):
        print("{}{} not found, please create one.{}".format(zcolors.WARNING, config_file, zcolors.ENDC))
        Exception('Config not found.')

    conf = configparser.ConfigParser()
    conf.read(config_file)

    return conf['xmm']


def check_if_not_create(file, template):
    """
    Checks for a file, if it doesn't exist, it will be created from a template.

    :param file:
        filename with path to file
    :type file: ``str``

    :param template:
        filename with path to template file
    :type template: ``str``
    """
    if not os.path.isfile(file):
        os.makedirs(os.path.dirname(file), exist_ok=True)
        copyfile(template, file)


def create_if_not_exists(file, contents):
    """
    Checks for a file, if it doesn't exist, it will be created from a template.

    :param file:
        filename with path to file
    :type file: ``str``

    :param contents:
        string contents of the file being created
    :type contents: ``str``
    """
    if not os.path.isfile(file):
        os.makedirs(os.path.dirname(file), exist_ok=True)
        with open(file, 'w') as f:
            f.write(contents)


def file_is_empty(filename):
    """
    Checks to see if a file is empty

    :param filename:
        string filename
    :type filename: ``str``

    :returns: ``bool``
    """
    return os.stat(filename).st_size == 0


def replace_last(string, old, new):
    """
    Replace the last occurrence of a pattern in a string

    :param string:
        string
    :type string: ``str``

    :param old:
        string to find
    :type old: ``str``

    :param new:
        string to replace
    :type new: ``str``

    :returns: ``str``
    """
    return string[::-1].replace(old[::-1], new[::-1], 1)[::-1]


def hash_file(filename):
    """
    Returns the SHA-1 hash of the file passed into it

    :param filename:
        string filename
    :type filename: ``str``

    :returns: ``str``
    """

    # make a hash object
    h = hashlib.sha1()

    # open file for reading in binary mode
    with open(filename, 'rb') as file:

        # loop till the end of the file
        chunk = 0
        while chunk != b'':
            # read only 1024 bytes at a time
            chunk = file.read(1024)
            h.update(chunk)

    # return the hex representation of digest
    return h.hexdigest()


# http://stackoverflow.com/questions/3041986/python-command-line-yes-no-input
def query_yes_no(question, default="yes"):
    """
    Ask a yes/no question via raw_input() and return their answer.

    :param question:
        a string that is presented to the user.
    :type question: ``str``

    :param default:
        is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).
    :type default: ``str``

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


# http://stackoverflow.com/a/24030569
class ObjectEncoder(json.JSONEncoder):
    """
    JSONEncoder subclass that leverages an object's `__json__()` method,
    if available, to obtain its default JSON representation.
    """
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        if hasattr(obj, '__json__'):
            return obj.__json__()

        return json.JSONEncoder.default(self, obj)


class zcolors:
    """
    Terminal formatting.

    Options:

        * HEADER
        * INFO
        * SUCCESS
        * WARNING
        * FAIL
        * ENDC (end color)
        * BOLD
        * UNDERLINE

    >>> "{}eggs{}: {}spam{}".format(zcolors.INFO, zcolors.ENDC, zcolors.UNDERLINE, zcolors.ENDC)

    """
    HEADER = '\033[95m'
    INFO = '\033[94m'
    SUCCESS = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def cprint(string, style='INFO'):
    """
    Terminal formatting convenience function.

    :param string:
        A string to print.
    :type string: ``str``

    :param style:
        A style to print.

        Options:

            * HEADER
            * INFO
            * SUCCESS
            * WARNING
            * FAIL
            * ENDC (end color)
            * BOLD
            * UNDERLINE

    :type style: ``str``

    >>> cprint("Success", style='SUCCESS')

    """
    color = getattr(zcolors, style)
    print('{}{}{}'.format(color, string, zcolors.ENDC))



import json
import os
import re
import urllib.request
from urllib.error import URLError

from xmm.map import MapPackage

from xmm.exceptions import PackageLookupError
from xmm.exceptions import RepositoryLookupError
from xmm.exceptions import RepositoryUpdateError
from xmm.base import Base
from xmm.util import zcolors
from xmm.util import cprint
from xmm import util


class Collection(Base):
    """
    A *Collection* is a collection of *Repository* objects

    :returns object: *Collection*

    >>> from xmm.repository import Collection
    >>> from xmm.repository import Repository
    >>> repositories = Collection()
    >>> repository = Repository(name='default', download_url='http://dl.repo.url/',
    >>>                         api_data_url='http://api.repo.url/maps.json', api_data_file='~/.xmm/maps.json')
    >>> repositories.add_repository(repository)

    """
    def __init__(self):
        super().__init__()
        self.sources = []

    def __repr__(self):
        return str(vars(self))

    def __json__(self):
        return {
            'sources': self.sources,
        }

    def to_json(self):
        """
        :returns: A **JSON** encoded version of this object
        """
        return json.dumps(self, cls=util.ObjectEncoder)

    def list_repositories(self):
        """
        Prints a list of servers
        """
        for repo in self.conf['sources']:
            print(repo)

    def get_repository(self, repository_name):
        """
        :param repository_name:
        :type repository_name: ``str``

        :returns: A **Repository** object or false if name not found

        >>> from xmm.repository import Collection
        >>> from xmm.repository import Repository
        >>> repositories = Collection()
        >>> repository = Repository(name='default', download_url='http://dl.repo.url/',
        >>>                         api_data_url='http://api.repo.url/maps.json', api_data_file='~/.xmm/maps.json')
        >>> repositories.add_repository(repository)
        >>> print(repositories.get_repository('default'))
        """
        self.logger.info("Getting repository: {}".format(repository_name))

        repo = False
        for source in self.sources:
            if source.name == repository_name:
                repo = source

        if not repo:
            raise RepositoryLookupError

        return repo

    def add_repository(self, repository):
        """
        Add a *Repository* to the *Collection*

        :param repository:
        :type repository: ``Repository``

        >>> from xmm.repository import Collection
        >>> from xmm.repository import Repository
        >>> repositories = Collection()
        >>> repository = Repository(name='default', download_url='http://dl.repo.url/',
        >>>                         api_data_url='http://api.repo.url/maps.json', api_data_file='~/.xmm/maps.json')
        >>> repositories.add_repository(repository)
        >>> print(repositories.get_repository('default'))
        """
        self.sources.append(repository)

    def search_all(self, bsp_name=False, gametype=False, author=False, title=False, pk3_name=False, shasum=False, detail=None, highlight=False):
        """
        Searches all *Repository* objects in the *Collection* for maps matching criteria

        :param bsp_name:
            Search by bsp name
        :type bsp_name: ``str``

        :param gametype:
            Search by gametype
        :type gametype: ``str``

        :param author:
            Search by author
        :type author: ``str``

        :param title:
            Search by title
        :type title: ``str``

        :param pk3_name:
            Search by pk3_name
        :type pk3_name: ``str``

        :param shasum:
            Search by shasum
        :type shasum: ``str``

        :param detail:
            How much detail in the results, [short, None, long]
        :type detail: ``str``

        :param highlight:
            Whether to highlight the search string
        :type highlight: ``bool``

        >>> from xmm.repository import Collection
        >>> from xmm.repository import Repository
        >>> repositories = Collection()
        >>> repository = Repository(name='default', download_url='http://dl.repo.url/',
        >>>                         api_data_url='http://api.repo.url/maps.json', api_data_file='~/.xmm/maps.json')
        >>> repositories.add_repository(repository)
        >>> print(repositories.search_all(bsp_name='vinegar_v3'))
        """

        self.logger.info("Searching all repositories.")

        for repo in self.sources:
            repo.search_maps(bsp_name=bsp_name, gametype=gametype, author=author, title=title, pk3_name=pk3_name, shasum=shasum, detail=detail, highlight=highlight)

    def update_all(self):
        """
        Update the data for all *Repository* objects in the *Collection*

        >>> from xmm.repository import Collection
        >>> from xmm.repository import Repository
        >>> repositories = Collection()
        >>> repository = Repository(name='default', download_url='http://dl.repo.url/',
        >>>                         api_data_url='http://api.repo.url/maps.json', api_data_file='~/.xmm/maps.json')
        >>> repositories.update_all()
        """

        self.logger.info("Updating all sources.")

        for repo in self.sources:
            self.logger.debug("Updating source: {}".format(repo.name))
            repo.update_repo_data()

    def export_all_hash_index(self, filename=None):
        """
        :param filename:
            Name for the exported json file, default ``all-repos-maps.json.shasums``
        :type filename: ``str``

        :returns: False if fails

        >>> from xmm.repository import Collection
        >>> from xmm.repository import Repository
        >>> repositories = Collection()
        >>> repository = Repository(name='default', download_url='http://dl.repo.url/',
        >>>                         api_data_url='http://api.repo.url/maps.json', api_data_file='~/.xmm/maps.json')
        >>> repositories.export_all_hash_index()
        """
        if not filename:
            filename = 'all-repos-maps.json.shasums'

        self.logger.info("exporting shasums from all sources to file: {}".format(filename))

        data = []
        for repo in self.sources:
            lines = repo.get_hash_index()
            data.extend(lines)

        if data:
            try:
                with open(filename, 'w') as f:
                    f.write('\n'.join(data))
            except EnvironmentError as e:
                self.logger.error(e)
                return False

    def export_all_packages(self, filename=None):
        """
        :param filename:
            Name for the exported json file, default ``maps.json``
        :type filename: ``str``

        :returns: False if fails

        >>> from xmm.repository import Collection
        >>> from xmm.repository import Repository
        >>> repositories = Collection()
        >>> repository = Repository(name='default', download_url='http://dl.repo.url/',
        >>>                         api_data_url='http://api.repo.url/maps.json', api_data_file='~/.xmm/maps.json')
        >>> repositories.export_all_packages()
        """
        if not filename:
            filename = 'all-repos-maps.json'

        self.logger.info("exporting maps as: {}".format(filename))

        data = []
        for repo in self.sources:
            lines = repo.get_packages()
            data.extend(lines)

        if data:
            try:
                with open(filename, 'w') as f:
                    f.write(json.dumps(data, cls=util.ObjectEncoder))
            except EnvironmentError as e:
                self.logger.error(e)
                return False


class Repository(Base):
    """
    A *Repository* contains a url which hosts content matching the
    **JSON** format described in the documentation

    :param name:
        A name for this *Repository*
    :type name: ``str``

    :param download_url:
        The url where the pk3 files should be downloaded from
    :type download_url: ``str``

    :param api_data_url:
        URL serving maps in the *JSON* format described in the documentation
    :type api_data_url: ``str``

    :param api_data_file:
        Local cache fo the repo data.
    :type api_data_file: ``str``

    :returns object: ``Repository``


    >>> from xmm.repository import Repository
    >>> repository = Repository(name='default', download_url='http://dl.repo.url/',
    >>>                         api_data_url='http://api.repo.url/maps.json', api_data_file='~/.xmm/maps.json')
    """
    def __init__(self, name, download_url, api_data_url, api_data_file):
        super().__init__()
        self.name = name
        self.api_data_url = api_data_url
        self.download_url = download_url
        self.api_data = None
        self.api_data_file = os.path.expanduser(api_data_file)
        self.repo_data = {}

    def __repr__(self):
        return str(vars(self))

    def __json__(self):
        return {
            'name': self.name,
            'api_data_url': self.api_data_url,
            'download_url': self.download_url,
            'api_data_file': self.api_data_file,
        }

    def to_json(self):
        """
        :returns: A **JSON** encoded version of this object
        """
        return json.dumps(self, cls=util.ObjectEncoder)

    def search_maps(self, bsp_name=False, gametype=False, author=False, title=False, pk3_name=False, shasum=False, detail=None, highlight=False):
        """
        Searches the repository for maps matching criteria

        :param bsp_name:
            Search by bsp name
        :type bsp_name: ``str``

        :param gametype:
            Search by gametype
        :type gametype: ``str``

        :param author:
            Search by author
        :type author: ``str``

        :param title:
            Search by title
        :type title: ``str``

        :param pk3_name:
            Search by pk3_name
        :type pk3_name: ``str``

        :param shasum:
            Search by shasum
        :type shasum: ``str``

        :param detail:
            How much detail in the results, [short, None, long]
        :type detail: ``str``

        :param highlight:
            Whether to highlight the search string
        :type highlight: ``bool``

        >>> from xmm.repository import Repository
        >>> repository = Repository(name='default', download_url='http://dl.repo.url/',
        >>>                         api_data_url='http://api.repo.url/maps.json', api_data_file='~/.xmm/maps.json')
        >>> repository.search_maps(bsp_name='dance' gametype='ctf')
        """

        self.logger.info("Searching maps.")

        maps_json = self.get_packages()
        fmaps_json = []
        criteria = []

        if not bsp_name:
            bsp_name = ''

        # Filter based on args
        total = 0
        for m in maps_json:

            show = False
            for bsp in m.bsp:
                if bsp_name and re.search('^.*' + bsp_name + '.*$', bsp):
                    criteria.append(('bsp', bsp_name))
                    show = True

                if gametype in m.bsp[bsp]['gametypes']:
                    criteria.append(('gametype', gametype))
                    show = True

                if re.search('^.*' + str(author) + '.*$', m.bsp[bsp]['author']):
                    criteria.append(('author', author))
                    show = True

                if re.search('^.*' + str(title) + '.*$', m.bsp[bsp]['title']):
                    criteria.append(('title', title))
                    show = True

            if pk3_name and re.search('^.*' + pk3_name + '.*$', m.pk3_file):
                criteria.append(('pk3', pk3_name))
                show = True

            if shasum == m.shasum:
                criteria.append(('shasum', shasum))
                show = True

            if show:
                total += 1
                fmaps_json.append(m)

        criteria = list(set(criteria))

        if len(criteria) > 0:
            cprint("Using repo '{}'".format(self.name), style="HEADER")
            cprint("Searching for packages with the following criteria:", style='INFO')
            for c in criteria:
                print("{}{}{}: {}".format(zcolors.BOLD, str(c[0]), zcolors.ENDC, str(c[1])))
            print('---')

        for m in fmaps_json:
            bsps = m.bsp
            keys = list(bsps)
            keys.sort()

            shown = False
            for bsp in keys:
                if re.search('^.*' + bsp_name + '.*$', bsp) and not shown:
                    if bsp_name:
                        m.show_map_details(search_string=bsp_name, detail=detail, highlight=highlight)
                    elif pk3_name != '':
                        m.show_map_details(search_string=pk3_name, detail=detail, highlight=highlight)
                    else:
                        m.show_map_details(detail=detail, highlight=highlight)

                    shown = True
        print('---')
        print("{}Total packages found:{} {}{}{}".format(zcolors.INFO, zcolors.ENDC, zcolors.BOLD, str(total), zcolors.ENDC))

    # remote data
    def update_repo_data(self):
        """
        Updates sources cache with latest maps from *Repository*

        >>> from xmm.repository import Repository
        >>> repository = Repository(name='default', download_url='http://dl.repo.url/',
        >>>                         api_data_url='http://api.repo.url/maps.json', api_data_file='~/.xmm/maps.json')
        >>> repository.update_repo_data()
        """
        try:
            cprint("Updating {} sources json...".format(self.name), style='INFO')
            self.logger.info("Updating {} sources json...".format(self.name))
            util.download_file(self.api_data_file, url=self.api_data_url, use_curl=self.conf['default']['use_curl'], overwrite=True)
        except URLError as e:
            self.logger.debug('Error updating repo data: {}'.format(e))
            raise RepositoryUpdateError

    def get_packages(self):
        """
        Gets the cached map list from *Repository* or reads from file if cache not available

        :returns: ``json``

        >>> from xmm.repository import Repository
        >>> repository = Repository(name='default', download_url='http://dl.repo.url/',
        >>>                         api_data_url='http://api.repo.url/maps.json', api_data_file='~/.xmm/maps.json')
        >>> print(repository.get_packages())
        """

        self.logger.debug("getting repo data")

        if not self.repo_data:

            repo_data = []

            if not os.path.exists(self.api_data_file):
                cprint("Could not find a repo file. Using maplist shipped with release. For the latest maps, run xmm update.".format(self.name), style='WARNING')
                self.logger.info("Could not find a repo file. Using maplist shipped with release. For the latest maps, run xmm update.".format(self.name))

                import zipfile
                zip_ref = zipfile.ZipFile(self.conf['default']['api_data_file_seed'], 'r')
                zip_ref.extract('maps.json', os.path.dirname(self.api_data_file))
                zip_ref.close()

            with open(self.api_data_file) as f:
                data = f.read()
                maps = json.loads(data)['data']

            for m in maps:
                new_map = MapPackage(map_package_json=m)
                repo_data.append(new_map)

            self.repo_data = repo_data

        return self.repo_data

    def export_packages(self, filename=None):
        """
        :param filename:
            Name for the exported json file, default ``maps.json``
        :type filename: ``str``

        :returns: False if fails

        >>> from xmm.repository import Repository
        >>> repository = Repository(name='default', download_url='http://dl.repo.url/',
        >>>                         api_data_url='http://api.repo.url/maps.json', api_data_file='~/.xmm/maps.json')
        >>> repository.export_packages('test.json')
        """
        if not filename:
            filename = 'xmm-export.maps.json'

        self.logger.info("exporting maps as: {}".format(filename))

        data = self.get_packages()

        if data:
            try:
                with open(filename, 'w') as f:
                    f.write(json.dumps(data, cls=util.ObjectEncoder))
            except EnvironmentError as e:
                self.logger.error(e)
                return False

    def get_hash_index(self):
        """
        Gets a list of all pk3s and their shasums

        :returns: False if fails

        >>> from xmm.repository import Repository
        >>> repository = Repository(name='default', download_url='http://dl.repo.url/',
        >>>                         api_data_url='http://api.repo.url/maps.json', api_data_file='~/.xmm/maps.json')
        >>> print(repository.get_hash_index())
        """
        maps_json = self.get_packages()
        lines = []
        for m in maps_json:
            lines.append("{} {}".format(m.shasum, m.pk3_file))

        return lines

    def export_hash_index(self, filename=None):
        """
        :param filename:
            Name for the exported json file, default ``maps.json.shasums``
        :type filename: ``str``

        :returns: False if fails

        >>> from xmm.repository import Repository
        >>> repository = Repository(name='default', download_url='http://dl.repo.url/',
        >>>                         api_data_url='http://api.repo.url/maps.json', api_data_file='~/.xmm/maps.json')
        >>> repository.export_hash_index('test.shasums')
        """
        if not filename:
            filename = 'xmm-export.maps.shasums'

        self.logger.info("exporting shasums to file: {}".format(filename))

        lines = self.get_hash_index()

        if lines:
            try:
                with open(filename, 'w') as f:
                    f.write('\n'.join(lines))
            except EnvironmentError as e:
                self.logger.error(e)
                return False

    def show_map(self, pk3_name, detail=None, highlight=False):
        """
        Convenience function to use the show_map_details helper

        :param pk3_name:
            The name of a pk3, such as ``vinegar_v3.pk3``
        :type pk3_name: ``str``

        :param detail:
            How much detail to show, [short, None, long]
        :type detail: ``str``

        :param highlight:
            Whether to highlight the results
        :type highlight: ``bool``

        :returns: ``MapPackage``

        >>> from xmm.repository import Repository
        >>> repository = Repository(name='default', download_url='http://dl.repo.url/',
        >>>                         api_data_url='http://api.repo.url/maps.json', api_data_file='~/.xmm/maps.json')
        >>> repository.show_map(pk3_name='vinegar_v3.pk3')
        """

        self.logger.debug("Showing map with helper")

        packages = self.get_packages()
        found_map = False

        for p in packages:
            if p.pk3_file == pk3_name:
                p.show_map_details(search_string=pk3_name, detail=detail, highlight=highlight)
                found_map = p

        if not found_map:
            raise PackageLookupError

        return found_map



import json

from xmm.map import MapPackage

from xmm.base import Base
from xmm import util


class Store(Base):
    """
    *Store* is for interacting with the datastore for a *Library*

    :param package_store_file:
        The file where the data is stored
    :type package_store_file: ``str``

    >>> import os
    >>> from xmm.store import Store
    >>> package_store_file = os.path.expanduser('~/.xmm/library.json')
    >>> store = Store(package_store_file=package_store_file)

    :returns object: ``Store``

    """
    def __init__(self, package_store_file):
        super().__init__()

        if package_store_file:
            util.create_if_not_exists(package_store_file, json.dumps([]))

        self.data_file = package_store_file
        self.data = self.get_package_db()

    def __repr__(self):
        return str(vars(self))

    def __json__(self):
        return {
            'data': self.data,
            'data_file': self.data_file,
        }

    def to_json(self):
        """
        :returns: A **JSON** encoded version of this object
        """
        return json.dumps(self, cls=util.ObjectEncoder)

    def get_package_db(self):
        """
        Searches the repository for maps matching criteria

        >>> import os
        >>> from xmm.store import Store
        >>> package_store_file = os.path.expanduser('~/.xmm/library.json')
        >>> store = Store(package_store_file=package_store_file)
        >>> store.get_package_db()

        :returns: ``dict``
        """

        self.logger.debug('Getting package db')

        package_data = []
        repo_data = []

        util.create_if_not_exists(self.data_file, json.dumps(package_data))

        if not util.file_is_empty(self.data_file):
            repo_data = []

            with open(self.data_file) as f:
                data = f.read()
                package_data = json.loads(data)

            for m in package_data:
                new_map = MapPackage(map_package_json=m)
                repo_data.append(new_map)

        return repo_data

    def add_package(self, package):
        """
        Adds a *MapPackage* to the *Library* *Store*

        :param package:
            MapPackage to add
        :type package: ``MapPackage``

        >>> import os
        >>> from xmm.map import MapPackage
        >>> from xmm.store import Store
        >>> package_store_file = os.path.expanduser('~/.xmm/library.json')
        >>> with open('my_map.json') as f:
        >>>     data = f.read()
        >>>     my_map = MapPackage(map_package_json=data)
        >>> store = Store(package_store_file=package_store_file)
        >>> store.add_package(my_map)


        :returns: False if fails
        """

        self.logger.info('Adding package: pk3={pk3}, filesize={filesize}, date={date}, shasum={shasum}'
                         .format(pk3=package.pk3_file,
                                 filesize=util.convert_size(package.filesize),
                                 date=package.date,
                                 shasum=package.shasum,
                                 )
                         )

        package_data = self.data
        package_data.append(package)

        # fix this
        data_out = []
        for m in package_data:
            data_out.append(json.loads(m.to_json()))

        try:
            with open(self.data_file, 'w+') as f:
                json.dump(data_out, f)
        except EnvironmentError as e:
            self.logger.error(e)
            return False

    def remove_package(self, package):
        """
        Removes a *MapPackage* to the *Library* *Store*

        :param package:
            MapPackage to remove
        :type package: ``MapPackage``

        >>> import os
        >>> from xmm.map import MapPackage
        >>> from xmm.store import Store
        >>> package_store_file = os.path.expanduser('~/.xmm/library.json')
        >>> with open('my_map.json') as f:
        >>>     data = f.read()
        >>>     my_map = MapPackage(map_package_json=data)
        >>> store = Store(package_store_file=package_store_file)
        >>> store.remove_package(my_map)

        :returns: False if fails
        """

        self.logger.info('Removing package: pk3={pk3}, filesize={filesize}, date={date}, shasum={shasum}'
                         .format(pk3=package.pk3_file,
                                 filesize=util.convert_size(package.filesize),
                                 date=package.date,
                                 shasum=package.shasum,
                                 )
                         )

        package_store = []

        if not util.file_is_empty(self.data_file):
            with open(self.data_file, 'r+') as f:
                package_store = json.load(f)
                package_store[:] = [m for m in package_store if (m.get('shasum') != package.shasum and m.get('pk3') != package.pk3_file)]

        try:
            with open(self.data_file, 'w+') as f:
                json.dump(package_store, f)
        except EnvironmentError as e:
            self.logger.error(e)
            return False

    def export_packages(self, filename=None):
        """
        Exports all *MapPackage* objects from the *Library* *Store*

        :param filename:
            Name for the exported json file, default ``xmm-export.json``
        :type filename: ``str``

        :returns: False if fails

        >>> from xmm.server import LocalServer
        >>> # Setup the store automatically with an instance of *LocalServer*
        >>> server = LocalServer()
        >>> server.library.store.export_packages(filename='test.json')
        """
        if not filename:
            filename = 'xmm-export.maps.json'

        data_out = []
        for m in self.data:
            data_out.append(json.loads(m.to_json()))

        self.logger.info('exporting maps as: {}'.format(filename))

        try:
            with open(filename, 'w') as f:
                f.write(json.dumps(data_out))
        except EnvironmentError as e:
            self.logger.error(e)
            return False



import logging


class ClassPrefixAdapter(logging.LoggerAdapter):
    def __init__(self, prefix, logger, extra=None):
        super(ClassPrefixAdapter, self).__init__(logger, extra or {})
        self.prefix = prefix
    """
    Custom Adapter to insert class name to provide context
    """
    def process(self, msg, kwargs):
        return '[%s] %s' % (self.prefix, msg), kwargs



from xmm.plugins import pluginbase
from xmm.util import zcolors
from xmm.util import cprint

config = pluginbase.get_config()


def get_args():
    command='hello'
    command_help={'help': 'hello is an example plugin'}
    args=['-f', '--foo']
    kwargs={'type': int, 'nargs': '?', 'help': 'this is a help line'}
    return command, command_help, args, kwargs


def run():
    print("Hello from a plugin!")
    cprint("I share the xmm util module", style='HEADER')
    cprint("plugins can use these functions and classes", style='BOLD')
    cprint("I share the xmm util module", style='INFO')
    cprint("oh noes! Something went wrong! (not really)", style='FAIL')
    cprint("It's okay! Don't Cry", style='WARNING')
    cprint("Get Serious", style='UNDERLINE')
    print("{}Look, I also have access to the config: {}".format(zcolors.SUCCESS, config['servers_config']))



from setuptools import setup
from setuptools import find_packages
from xmm import __author__, __email__, __url__, __license__, __version__, __summary__, __keywords__


with open('requirements.in') as f:
    install_requires = f.read().splitlines()

try:
    import pypandoc
    readme_contents = pypandoc.convert('README.md', 'rst')
    changelog_contents = pypandoc.convert('CHANGELOG.md', 'rst')
except(IOError, ImportError):
    with open('README.md') as f:
        readme_contents = f.read()
    with open('CHANGELOG.md') as f:
        changelog_contents = f.read()

long_description = '{}\n{}'.format(readme_contents, changelog_contents)

setup(
    name='xmm',
    version=__version__,
    description=__summary__,
    long_description=long_description,
    author=__author__,
    author_email=__email__,
    url=__url__,
    license=__license__,
    packages=find_packages(exclude=('tests', 'docs')),
    package_data={'': ['LICENSE', 'README.md', 'CHANGELOG.md', 'docs/*', 'config/*', 'bin/*', 'pkg/*', 'resources/*']},
    include_package_data=True,
    install_requires=install_requires,
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    scripts=['pkg/xmm.bash', 'pkg/xmm.zsh'],
    entry_points={
       'console_scripts': [
          'xmm = xmm.cli:main'
       ]
    },
    keywords=__keywords__,
    classifiers=[
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: System Administrators',

        'License :: OSI Approved :: MIT License',

        'Development Status :: 5 - Production/Stable',

        'Environment :: Console',

        'Topic :: Games/Entertainment',
        'Topic :: System :: Archiving :: Packaging',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)



import time
import json

from xmm.base import Base
from xmm.util import zcolors
from xmm import util


class MapPackage(Base):
    """
    *MapPackage* contains top-level metadata about a pk3 file and list of *Bsp* objects inside this package

    :param map_package_json:
        A dict or *JSON* string that matches "specification" in the Developers section of the documentation.

        See basic example below:

        .. code-block:: json

            {
              "data": [
                {
                  "date": 1453749340,
                  "filesize": 7856907,
                  "bsp": {
                    "vapor_alpha_2": {
                      "radar": "gfx/vapor_alpha_2_mini.tga",
                      "waypoints": "",
                      "title": "Vapor",
                      "description": "Such CTF. Many Vehicles. Wow.",
                      "map": "maps/vapor_alpha_2.map",
                      "entities": {
                        "info_player_deathmatch": 4,
                        "info_player_team1": 11,
                        "info_player_team2": 11,
                        "item_armor_big": 10,
                        "item_armor_large": 4,
                        "item_armor_medium": 16,
                        "item_armor_small": 124,
                        "item_bullets": 10,
                        "item_cells": 14,
                        "item_flag_team1": 1,
                        "item_flag_team2": 1,
                        "item_health_large": 6,
                        "item_health_medium": 30,
                        "item_health_mega": 2,
                        "item_health_small": 100,
                        "item_invincible": 1,
                        "item_rockets": 20,
                        "item_strength": 1,
                        "weapon_crylink": 4,
                        "weapon_devastator": 6,
                        "weapon_electro": 2,
                        "weapon_grenadelauncher": 6,
                        "weapon_hagar": 4,
                        "weapon_machinegun": 6,
                        "weapon_vortex": 4
                      },
                      "mapinfo": "maps/vapor_alpha_2.mapinfo",
                      "author": "-z-",
                      "gametypes": [
                        "ctf",
                        "DM"
                      ],
                      "license": true,
                      "mapshot": "maps/vapor_alpha_2.jpg"
                    }
                  },
                  "shasum": "3df0143516f72269f465070373f165c8787964d5",
                  "pk3": "map-vapor_alpha_2.pk3"
                }
              ]
            }

    :type map_package_json: ``string|dict``

    :returns object: ``MapPackage``

    >>> from xmm.map import MapPackage
    >>> with open('my_map.json') as f:
    >>>     data = f.read()
    >>>     my_map = MapPackage(map_package_json=data)

    """
    def __init__(self, map_package_json):
        super().__init__()

        if not isinstance(map_package_json, dict):
            map_package = json.loads(map_package_json)
        else:
            map_package = map_package_json

        self.pk3_file = map_package['pk3']
        self.shasum = map_package['shasum']
        self._bsp = map_package['bsp']
        self.date = map_package['date']
        self.filesize = map_package['filesize']

    @property
    def bsp(self):
        return self._bsp

    @bsp.setter
    def bsp(self, key, value):
        self._bsp[key] = value

    @bsp.deleter
    def bsp(self, key):
        del self._bsp[key]

    def __repr__(self):
        return 'MapPackage(pk3=%s, shasum=%s, bsp=%s, date=%s, filesize=%s)' % (self.pk3_file, self.shasum, repr(self.bsp), self.date, self.filesize)

    def __json__(self):
        return {
            'pk3': self.pk3_file,
            'shasum': self.shasum,
            'filesize': self.filesize,
            'date': self.date,
            'bsp': self.bsp,
        }

    def to_json(self):
        """
        :returns: A **JSON** encoded version of this object
        """
        return json.dumps(self, cls=util.ObjectEncoder)

    def show_map_details(self, detail=None, search_string='', highlight=False):
        """
        Helper function for pretty printing details about a *MapPackage*

        Convenience function to use the show_map_details helper

        :param detail:
            How much detail to show, [short, None, long]
        :type detail: ``str``

        :param search_string:
            A string to highlight with ``highlight=True``
        :type search_string: ``str``

        :param highlight:
            Whether to highlight the results
        :type highlight: ``bool``

        :returns: ``MapPackage``
        """

        self.logger.debug('Showing details for map: {}'.format(self.pk3_file))

        bsps = self.bsp
        keys = list(bsps)
        keys.sort()
        search_string = str(search_string)

        # Long view
        if detail == 'long':
            print('')
            print('         pk3: {}{}{}'.format(zcolors.BOLD, str(self.pk3_file), zcolors.ENDC))

            for bsp in keys:
                # Handle highlight
                string_highlighted = '{}{}{}{}{}'.format(zcolors.ENDC, zcolors.SUCCESS, search_string, zcolors.ENDC, zcolors.INFO)
                bsp_highlighted = bsp.replace(search_string, string_highlighted)
                if search_string and highlight:
                    print('         bsp: {}{}{}'.format(zcolors.INFO, bsp_highlighted, zcolors.ENDC))
                else:
                    print('         bsp: {}{}{}'.format(zcolors.INFO, bsp, zcolors.ENDC))

                # bsp specific
                print('       title:  {}'.format(str(self.bsp[bsp]['title'])))
                print(' description:  {}'.format(str(self.bsp[bsp]['description'])))
                print('      author:  {}'.format(str(self.bsp[bsp]['author'])))

            # pk3 specific
            print('      shasum: {}'.format(str(self.shasum)))
            print('      shasum: {}'.format(str(self.pk3_file)))
            print('        date: {}'.format(time.strftime('%Y-%m-%d', time.localtime(self.date))))
            print('        size: {}'.format(util.convert_size(self.filesize).strip()))
            print('          dl: {}'.format(self.conf['default']['download_url'] + self.pk3_file))

        # Short detail view
        elif detail == 'short':
            print(str(self.pk3_file))
        # Default view
        else:
            bsp_string = '\n{}{}{} ['.format(zcolors.BOLD, str(self.pk3_file), zcolors.ENDC)
            for bsp in keys:
                # Handle highlight
                string_highlighted = '{}{}{}{}{}'.format(zcolors.ENDC, zcolors.SUCCESS, search_string, zcolors.ENDC, zcolors.INFO)
                bsp_highlighted = bsp.replace(search_string, string_highlighted)
                if search_string and highlight:
                    bsp_string += '{}{}{}, '.format(zcolors.INFO, bsp_highlighted, zcolors.ENDC)
                else:
                    bsp_string += '{}{}{}, '.format(zcolors.INFO, bsp, zcolors.ENDC)
            bsp_string = util.replace_last(bsp_string, ', ', '')
            bsp_string += ']'.format(zcolors.ENDC)
            print(bsp_string)
            print('{}{}'.format(self.conf['default']['download_url'], str(self.pk3_file)))


class Bsp(object):
    """
    A *Bsp* is a child of a *MapPackage* that holds metadata about this map

    :param pk3_file:
        The pk3_file name of the package this bsp is in
    :type pk3_file: ``str``

    :param bsp_name:
        The bsp_name of the bsp_file
    :type bsp_name: ``str``

    :param bsp_file:
        The bsp_file
    :type bsp_file: ``str``

    :param map_file:
        The map_file for the bsp_file if it exists
    :type map_file: ``str``

    :param mapshot:
        The mapshot for the bsp_file if it exists
    :type mapshot: ``str``

    :param title:
        The title for the bsp_file if it exists
    :type title: ``str``

    :param description:
        The description for the bsp_file if it exists
    :type description: ``str``

    :param mapinfo:
        The mapinfo for the bsp_file if it exists
    :type mapinfo: ``str``

    :param author:
        The author for the bsp_file if it exists
    :type author: ``str``

    :param gametypes:
        The gametypes for the bsp_file if it exists
    :type gametypes: ``list``

    :param entities:
        The entities for the bsp_file if they exists
    :type entities: ``str``

    :param waypoints:
        The waypoints for the bsp_file if it exists
    :type waypoints: ``str``

    :param license:
        The license for the bsp_file if it exists
    :type license: ``str``

    :returns object: ``Bsp``

    """
    def __init__(self, pk3_file='', bsp_name='', bsp_file='', map_file='', mapshot='', radar='', title='', description='', mapinfo='', author='', gametypes=None, entities=None, waypoints='', license=False):
        self.pk3_file = pk3_file
        self.bsp_name = bsp_name
        self.bsp_file = bsp_file
        self.map_file = map_file
        self.mapshot = mapshot
        self.radar = radar
        self.title = title
        self.description = description
        self.mapinfo = mapinfo
        self.author = author
        self.gametypes = gametypes
        self.entities = entities
        self.waypoints = waypoints
        self.license = license

    def __repr__(self):
        return 'Bsp(pk3_file=%s, bsp_name=%s, bsp_file=%s, map_file=%s, mapshot=%s, radar=%s, title=%s, description=%s, mapinfo=%s, author=%s, gametypes=%s,entities=%s, waypoints=%s, license=%s)' % (self.pk3_file, self.bsp_name, self.bsp_file, self.map_file, self.mapshot, self.radar, self.title, self.description, self.mapinfo, self.author, self.gametypes, self.entities, self.waypoints, self.license)

    def __json__(self):
        return {
            'map': self.map_file,
            'mapshot': self.mapshot,
            'radar': self.radar,
            'title': self.title,
            'description': self.description,
            'mapinfo': self.mapinfo,
            'author': self.author,
            'gametypes': self.gametypes,
            'entities': self.entities,
            'waypoints': self.waypoints,
            'license': self.license,
        }

    def to_json(self):
        """
        :returns: A **JSON** encoded version of this object
        """
        return json.dumps(self, cls=util.ObjectEncoder)



import xmm.util as util
import json
import logging.config
import os

config_file = '.xmm.ini'
logging_config_file = 'xmm.logging.ini'
config_file_with_path = os.path.expanduser(os.path.join('~', config_file))
logging_config_file_with_path = os.path.expanduser(os.path.join('~/.xmm/', logging_config_file))

util.check_if_not_create(config_file_with_path, 'config/xmm.ini')
util.check_if_not_create(logging_config_file_with_path, 'config/xmm.logging.ini')

config = util.parse_config(config_file_with_path)

conf = {
    'default': {
        'library': os.path.expanduser('~/.xmm/library.json'),
        'target_dir': os.path.expanduser(config['target_dir']),
        'download_url': os.path.expanduser(config['download_url']),
        'api_data_file': os.path.expanduser(config['api_data_file']),
        'api_data_url': os.path.expanduser(config['api_data_url']),
        'api_data_file_seed': os.path.expanduser('~/.xmm/maps.json.zip'),
        'use_curl': False,
    },
    'sources_config': os.path.expanduser(config['sources_config']),
    'servers_config': os.path.expanduser(config['servers_config']),
    'sources': {},
    'servers': {},
}

# Overcome ini pitfalls
if config['use_curl'].lower() == 'true':
    conf['default']['use_curl'] = True

util.check_if_not_create(conf['sources_config'], 'config/example.sources.json')
util.check_if_not_create(conf['servers_config'], 'config/example.servers.json')

# If maplist doesn't exist copy seed zip over
util.check_if_not_create(conf['default']['api_data_file_seed'], 'resources/data/maps.json.zip')

# Load JSON files are dicts
with open(conf['sources_config']) as f:
    data = f.read()
    sources = json.loads(data)
    conf['sources'] = sources

with open(conf['servers_config']) as f:
    data = f.read()
    servers = json.loads(data)
    conf['servers'] = servers

# Make sure needed dirs exist
for server in conf['servers']:
    os.makedirs(os.path.expanduser(conf['servers'][server]['target_dir']), exist_ok=True)

logging.config.fileConfig(logging_config_file_with_path, defaults={
    'log_filename': os.path.expanduser('~/.xmm/xmm.log')
})
